!sprint
!2345678911234567892123456789312345678941234567895123456789612345678971234567898
!        0         0         0         0         0         0         0         0
!
! Module: evproject_mod
!
! Authors:  Andy Read - XMM-SSC-MPE
!           Michael Freyberg - MPE
!           Ian Stewart - SSC/LUX
!
! Description: a
! 
! Comments: 
! 
! $Id: module.f90,v 1.5 1998/01/13 08:39:15 gvacanti Exp $
!- 
module evproject_mod

  use cal
  use oal
  use dal
  use param  
  use errorHandling
!  use emutils_mod

  use binned_att

use array_utils
  use constants
  use coordinate
  use ssc_misc ! for ucase
  use wcs_aux

  implicit none 
  
  private 

  character(256) ::&
    binnedAttSetName='',&
    pntKeywordSetName=''
  character(20) ::&
    sourceOfAttInfo='',&
    sourceOfOdfAttInfo='',&
    xColName='',&
    yColName='',&
    tanRayStyle=''
  logical ::&
    writeBinnedAttPlease=.false.,&
    useMedianForPnt=.false.
  real(double) ::&
    maxDeltaAngleArcsec=0.0d0

  public evproject

contains 

  !---------------------------------------------------------------------
  subroutine evproject
  !---------------------------------------------------------------------

    type(DataSetT) :: eventSet, binnedAttSet, pntSet
    type(TableT) :: eventTab, binnedAttTab
    integer ::&
      numEvents, numTimeChanges, numBadEvents, numOutOfRange,&
      rowOfFirstGoodAtt, row, numBins, value
    type(ColumnT) :: timeECol, detxCol, detyCol, xCol, yCol
    real(double), pointer :: p_timeE(:), tanThetaArcsec(:), phi(:)
    integer(int16), pointer :: p_detx(:), p_dety(:)
    character(7) :: instrumentStr, datamode, pntStyle
    integer(int32) :: instrumentId
    real(double) ::&
      startTimeFirstBin, projRaDeg, projDecDeg,&
      maxDeltaAngle, tanX, tanY, skyPixSizeDeg
!, pixsize_x, pixsize_y
    integer(int32), pointer :: p_x(:), p_y(:)
    real(single) :: imagesize, x, y, Xmin, Xmax, Ymin, Ymax
    real(single), parameter ::&
      skyPixSizeArcsec = 0.05, nominalFovRadiusArcmin = 15.0
    type(attBinType), pointer :: binnedAtt(:)
    logical :: endsUnbounded(2), timechange, attIsGood, attIsNew
    type(SpacecraftAttitudeType) :: scAttitude, pntDeg, maxDelta, attLo, attHi
    type(WcsT) :: xyColWcs
    character(256) :: history
logical :: testFlag = .false.

!******************************************************************************
!	START 
!
!------------------------------------------
!       OPEN EVENT LIST
!------------------------------------------
    eventSet = dataset(stringParameter("eventset"),MODIFY)
    if (.not.hasBlock(eventSet,"EVENTS")) then
call error("noEventsExtension","Input file incorrect - no EVENTS extension")
    endif
    eventTab = table(eventSet,"EVENTS")
    numEvents = numberOfRows(eventTab)

    if (numEvents <= 0) then
!call error('noEvents',"")
      call warning('noEvents',"")
      call release(eventTab)
      call release(eventSet)
  return
    end if

    if (.not.hasColumn(eventTab,"TIME")) then
call error("noTimeColumn","Input file incorrect - no TIME column")
    endif
    if (.not.hasColumn(eventTab,"DETX")) then
call error("noDetxColumn","Input file incorrect - no DETX column")
    endif
    if (.not.hasColumn(eventTab,"DETY")) then
call error("noDetyColumn","Input file incorrect - no DETY column")
    endif
    timeECol = column(eventTab,"TIME",READ)
    detxCol = column(eventTab,"DETX",READ)
    detyCol = column(eventTab,"DETY",READ)
    p_timeE => Real64Data(timeECol)
    p_detx => Int16Data(detxCol)
    p_dety => Int16Data(detyCol)
!------------------------------------------------
!	SET CAL/OAL STATE 
!------------------------------------------------
    instrumentStr = stringAttribute(eventSet, 'INSTRUME')
    instrumentId = instId(instrumentStr)
    call CAL_setState(instrument=instrumentId, randomize=.FALSE.)

    if (instrumentId /= EPN.and. &
        instrumentId /= EMOS1 .and. &
        instrumentId /= EMOS2) then 
!        instrument /= "RGS1" .and. &
!        instrument /= "RGS2") then
      write (errstr,*) "Instrument ( ",instrumentStr," ) not supported"
call error("badInstrument")
    endif
    write(errstr,*) "Instrument: ",instrumentStr
    call message(AppMsg,Verbose)
!------------------------------------------------
!	CONVERT DETXY TO TELCOORD 
!------------------------------------------------
    allocate(tanThetaArcsec(0:numEvents-1), phi(0:numEvents-1))
    call detToTel(real(p_detx, double), real(p_dety, double)&
      , tanThetaArcsec, phi) ! in ssclib/coordinate.

!print *, ''
!do row = 0, numEvents-1
!print *, ''
!print *, 'row', row
!print *, '  p_detx, p_dety:', p_detx(row), p_dety(row)
!print *, '  theta,     phi:', tanThetaArcsec(row), phi(row)
!end do
!print *, ''

    call release(detxCol)
    call release(detyCol)
!-------------------------------------------
!     GET DATAMODE 
!-------------------------------------------
    datamode = ucase(stringAttribute(eventSet,"DATAMODE"))
    if (datamode /= "IMAGING" .and. datamode /= "TIMING"&
    .and. datamode /= "BURST") then
      write (errstr,*) "Mode ( ",datamode," ) not supported"
call error("badDatamode")
    endif
    write(errstr,*) "Datamode: ",datamode
    call message(AppMsg,Verbose)
!-------------------------------------------
!!     GET TCDLT keywords
!-------------------------------------------
!    pixsize_x = real64Attribute(detxCol,"TCDLT")
!    pixsize_y = real64Attribute(detyCol,"TCDLT")
!---------------------------------------------
!       ADD X/Y COLUMNS IF THEY DON'T EXIST (they should)
!---------------------------------------------
    xColName = stringParameter('xcol')
    yColName = stringParameter('ycol')
    if (.not.hasColumn(eventTab,xColName)) then 
      call message(AppMsg,Sparse," Column "//trim(xColName)&
       //" does not exist - creating")
!      dims(1)=1
      xCol = addColumn(eventTab,xColName,Integer32,"0.05 arcsec",(/1_int32/))
    else
      xCol = column(eventTab,xColName,MODIFY)
    endif
    if (.not.hasColumn(eventTab,yColName)) then 
      call message(AppMsg,Sparse," Column "//trim(yColName)&
       //" does not exist - creating")
!      dims(1)=1
      yCol = addColumn(eventTab,yColName,Integer32,"0.05 arcsec",(/1_int32/))
    else
      yCol = column(eventTab,yColName,MODIFY)
    endif

    p_x => Int32Data(xCol)
    p_y => Int32Data(yCol)
!------------------------------------------------- 
!       GET IMAGE SIZE (and convert to sky pixels) 
!-------------------------------------------------
    imagesize = realParameter("imagesize")   !NEW in degrees (def 0.36 = 21.6')
    if (imagesize /= 0.36) then 	     !different from default 
      write(errstr,*) "Note: imagesize chosen to be different from default &
        &value (0.36 deg)"
    endif
    imagesize = 3600.0 * imagesize / skyPixSizeArcsec
    ! in units of 0.05 arcseconds.

    write(errstr,*) "Image size [in 0.05 arcs.] : ",imagesize
    call message(AppMsg,Verbose)
!---------------------------------------------------
!	GET SOURCE OF ATTITUDE INFO (ATTSRC)
!---------------------------------------------------
    sourceOfAttInfo = ucase(stringParameter("attsource"))
    if (sourceOfAttInfo /= "FIXED" .and. &
        sourceOfAttInfo /= "BINNED" .and. &
        sourceOfAttInfo /= "ODF") then
      write (errstr,*) &
        "Source of attitude information ( ",sourceOfAttInfo," ) incorrect &
          &- Should be FIXED, BINNED or ODF"
call error("badAttitudeSource") 
    endif	
    write(errstr,*) "Source of attitude information: ",sourceOfAttInfo
    call message(AppMsg,Verbose)

    select case(sourceOfAttInfo)
    case('BINNED')
      binnedAttSetName = stringParameter('inbinnedattset')
      call readBinnedAttitude(binnedAttSetName, binnedAtt&
        , startTimeFirstBin, endsUnbounded)

    case('ODF')
      if (any(p_timeE(0:numEvents-2) > p_timeE(1:numEvents-1)))&
call error('badEventTimeOrder',"")

      sourceOfOdfAttInfo = ucase(stringParameter("odfattsource"))
      if (sourceOfOdfAttInfo == "AHF") then 
        call OAL_setState(attitudeFromAhf = .true.)       
      else if (sourceOfOdfAttInfo == "OM") then
        call OAL_setState(attitudeFromAhf = .false.)
      else
call error('badOdfAttitudeSource',"")
      endif

      writeBinnedAttPlease = booleanParameter('writebinnedatt')
      if (writeBinnedAttPlease) then
        allocate(binnedAtt(numEvents))
        binnedAtt%fracExp = 1.0
      end if

    case('FIXED')
      scAttitude%ra   = PiOn180 * realParameter("attra")  
      scAttitude%dec  = PiOn180 * realParameter("attdec")            
      scAttitude%apos = PiOn180 * realParameter("attapos")            

    case default
      write (errstr,*) &
        "Source of attitude information ( ",sourceOfAttInfo," ) incorrect &
          &- Should be FIXED, BINNED or ODF"
call error("badAttitudeSource") 
    end select
!-------------------------------------------------
!	GET PNT values 
!-------------------------------------------------
    select case(sourceOfAttInfo)
    case('BINNED')
      binnedAttSet = dataSet(binnedAttSetName, READ)
      binnedAttTab = table(binnedAttSet, 'ATT_BINS')
      pntDeg%ra   = real64Attribute(binnedAttTab, 'RA_PNT')
      pntDeg%dec  = real64Attribute(binnedAttTab, 'DEC_PNT')
      pntDeg%apos = real64Attribute(binnedAttTab, 'PA_PNT')
      pntStyle = stringAttribute(binnedAttTab, 'AVRG_PNT')
      call release(binnedAttSet)

    case('ODF')
      pntKeywordSetName = stringParameter('pntkwdset')
      pntSet = dataSet(pntKeywordSetName, READ)
      useMedianForPnt = booleanParameter('withmedianpnt')
      if (useMedianForPnt) then
        select case(sourceOfOdfAttInfo)
        case('AHF')
          pntDeg%ra   = real64Attribute(pntSet,"MAHFRA")
          pntDeg%dec  = real64Attribute(pntSet,"MAHFDEC")
          pntDeg%apos = real64Attribute(pntSet,"MAHFPA")
        case('OM')
          pntDeg%ra   = real64Attribute(pntSet,"MOMRA")
          pntDeg%dec  = real64Attribute(pntSet,"MOMDEC")
          pntDeg%apos = real64Attribute(pntSet,"MOMPA")
        case default
call error('badOdfAttitudeSource',"")
        end select

        pntStyle = 'MEDIAN'            

      else ! use the mean for pnt
        select case(sourceOfOdfAttInfo)
        case('AHF')
          pntDeg%ra   = real64Attribute(pntSet,"AAHFRA")
          pntDeg%dec  = real64Attribute(pntSet,"AAHFDEC")
          pntDeg%apos = real64Attribute(pntSet,"AAHFPA")
        case('OM')
          pntDeg%ra   = real64Attribute(pntSet,"AOMRA")
          pntDeg%dec  = real64Attribute(pntSet,"AOMDEC")
          pntDeg%apos = real64Attribute(pntSet,"AOMPA")
        case default
call error('badOdfAttitudeSource',"")
        end select

        pntStyle = 'MEAN'            
      end if
      call release(pntSet)

    case('FIXED')
      pntDeg%ra   = scAttitude%ra   / PiOn180
      pntDeg%dec  = scAttitude%dec  / PiOn180
      pntDeg%apos = scAttitude%apos / PiOn180
      pntStyle = 'FIXED'            

    case default
      write (errstr,*) &
        "Source of attitude information ( ",sourceOfAttInfo," ) incorrect &
          &- Should be FIXED, BINNED or ODF"
call error("badAttitudeSource") 
    end select
!-------------------------------------------------
!	GET DIRECTION OF PROJECTION-PLANE TANGENT (used to be ref point)
!-------------------------------------------------
    tanRayStyle = stringParameter('tangdirstyle')
    select case(tanRayStyle)
    case('user')
      projRaDeg  = realParameter("tangdirra")    
      projDecDeg = realParameter("tangdirdec")  

    case('nom')
      projRaDeg  = real64Attribute(eventSet,"RA_NOM")
      projDecDeg = real64Attribute(eventSet,"DEC_NOM")

    case('obj')
      projRaDeg  = real64Attribute(eventSet,"RA_OBJ")
      projDecDeg = real64Attribute(eventSet,"DEC_OBJ")

    case('pnt')
      projRaDeg  = pntDeg%ra
      projDecDeg = pntDeg%dec

    case default
      write (errstr,*) &
        "Source of tangent point information ( ",tanRayStyle," ) incorrect &
          &- Should be NOM, USER, PNT or OBJ"
call error("badTangentPoint") 
    end select
    write(errstr,*) "Source of tangent point information: ",tanRayStyle
    call message(AppMsg,Verbose)
    write(errstr,*) "Tangent point co-ords. ",projRaDeg,projDecDeg
    call message(AppMsg,Verbose)
!-------------------------------------------------
!	CALCULATE MAX ATT DELTA
!-------------------------------------------------
    if (sourceOfAttInfo=='ODF') then
      maxDeltaAngleArcsec = realParameter('maxdelta')
      maxDeltaAngle = PiOn180 * (maxDeltaAngleArcsec / 3600.0d0)
      maxDelta%ra   = maxDeltaAngle / cos(pntDeg%dec * PiOn180)
      maxDelta%dec  = maxDeltaAngle
      maxDelta%apos = maxDeltaAngle / (nominalFovRadiusArcmin * PiOn180 / 60.0)
    end if
!------------------------------------------------------
! 	MISC INITIALIZE
!------------------------------------------------------
    timechange = .TRUE.
    numTimeChanges = 1
    numBadEvents = 0
!    numBadAtts = 0
    numOutOfRange = 0
    attIsNew = .true. ! default
!-------------------------------------------
!	Set NULL values for X and Y
!-------------------------------------------
    call setNullValue(xCol,-99999999)
    call nullcheck(nullable(xCol))
    call setNullValue(yCol,-99999999)
    call nullcheck(nullable(yCol))
!-------------------------------------------
!	SETUP WCS FOR X,Y COLS
!-------------------------------------------
    xyColWcs%x%ctype         = 'RA---TAN'
    xyColWcs%x%worldUnit     = 'deg'
    xyColWcs%x%refPixels     = nint(imagesize + 1)
    xyColWcs%x%refWorld      = projRaDeg

    xyColWcs%y%ctype         = 'DEC--TAN'
    xyColWcs%y%worldUnit     = 'deg'
    xyColWcs%y%refPixels     = nint(imagesize + 1)
    xyColWcs%y%refWorld      = projDecDeg

!    call convertOldToNew(-pixsize_x, pixsize_y, 0.0d0, xyColWcs)
    skyPixSizeDeg = skyPixSizeArcsec / 3600.0d0
    call convertOldToNew(-skyPixSizeDeg, skyPixSizeDeg, 0.0d0, xyColWcs)
    ! in ssclib/wcs_aux

!------------------------------------------------------
!
!*****************  MAIN LOOP THROUGH EVENTS ***********
!
!------------------------------------------------------

    rowOfFirstGoodAtt = numEvents
    do row = 0, numEvents-1
      select case(sourceOfAttInfo)
      case('BINNED')
        call sampleBinnedAttitude(binnedAtt, startTimeFirstBin, p_timeE(row)&
          , endsUnbounded, scAttitude, attIsGood, attIsNew)

      case('ODF')
        call binUpAttitudeFromOdf(p_timeE(row), maxDelta, scAttitude&
          , attIsGood, attIsNew)
!print *, ''
!print *, 'row=', row
!print *, 'attIsGood=', attIsGood
!print *, 'attIsNew=', attIsNew
!print *, 'att%ra, att%dec, att%apos:', scAttitude%ra, scAttitude%dec, scAttitude%apos

        if (attIsNew .and. writeBinnedAttPlease) then
          numBins = 1
          startTimeFirstBin = p_timeE(row)
          binnedAtt(numBins)%att = scAttitude
          binnedAtt(numBins)%isGood = attIsGood
        end if
      end select ! default is to use fixed

      if (attIsGood) then
        rowOfFirstGoodAtt = row
    exit
      else
        numBadEvents = numBadEvents + 1
        call setToNull(xCol,row)
        call setToNull(yCol,row)
      end if
    end do
if (rowOfFirstGoodAtt >= numEvents) call error('noGoodAttitudes',"")

    row = rowOfFirstGoodAtt

    call telToTan(tanThetaArcsec(row), phi(row)&
      , xyColWcs%x%refWorld * PiOn180, xyColWcs%y%refWorld * PiOn180&
      , scAttitude, p_timeE(row), tanX, tanY)
    ! in ssclib/coordinate

    call tanToPixels(xyColWcs, tanX, tanY, x, y)
    ! in ssclib/wcs_aux

!print *, ''
!print *, 'theta, phi:', tanThetaArcsec(row), phi(row)
!print *, 'tanX, tanY:', tanX, tanY
!print *, '   x,    y:', x, y

    p_x(row) = nint(x)
    p_y(row) = nint(y)

    if ( x < 1 .or. x > imagesize*2&
    .or. y < 1 .or. y > imagesize*2) then
      numOutOfRange = numOutOfRange + 1
    endif

    attHi%ra   = scAttitude%ra
    attLo%ra   = scAttitude%ra
    attHi%dec  = scAttitude%dec
    attLo%dec  = scAttitude%dec
    attHi%apos = scAttitude%apos
    attLo%apos = scAttitude%apos
    Xmax = x
    Ymax = y
    Xmin = x
    Ymin = y

    numTimeChanges = 0
    attIsNew = .false. ! default
    do row = rowOfFirstGoodAtt + 1, numEvents-1
      errStr = sprintEvery(row, numEvents, fraction=0.1, message='  ')
      ! in ssclib/ssc_misc.
if (testFlag) print *, 'aaa'

      if (errStr .ne. '') call message(appMsg, verbose)

      if (abs(p_timeE(row) - p_timeE(row-1)) >= 0.0001) then 
        select case(sourceOfAttInfo)
        case('BINNED')
if (testFlag) print *, 'bbb'
          call sampleBinnedAttitude(binnedAtt, startTimeFirstBin, p_timeE(row)&
            , endsUnbounded, scAttitude, attIsGood, attIsNew)
if (testFlag) print *, 'ccc'

        case('ODF')
          call binUpAttitudeFromOdf(p_timeE(row), maxDelta, scAttitude&
            , attIsGood, attIsNew)
!print *, ''
!print *, 'row=', row
!print *, 'attIsGood=', attIsGood
!print *, 'attIsNew=', attIsNew
!print *, 'att%ra, att%dec, att%apos:', scAttitude%ra, scAttitude%dec, scAttitude%apos

          if (attIsNew .and. writeBinnedAttPlease) then
            numBins = numBins + 1
            binnedAtt(numBins-1)%binEndTime = p_timeE(row)
            binnedAtt(numBins)%att = scAttitude
            binnedAtt(numBins)%isGood = attIsGood
          end if
        end select ! default is to use fixed
      else 
        numTimeChanges = numTimeChanges + 1
      end if
if (testFlag) print *, 'ddd'

      if (.not. attIsGood) then 
        numBadEvents = numBadEvents + 1
        call setToNull(xCol,row)
        call setToNull(yCol,row)

      else ! attIsGood
        call telToTan(tanThetaArcsec(row), phi(row)&
          , xyColWcs%x%refWorld * PiOn180, xyColWcs%y%refWorld * PiOn180&
          , scAttitude, p_timeE(row), tanX, tanY, logical(attIsNew, bool))
        ! in ssclib/coordinate

        call tanToPixels(xyColWcs, tanX, tanY, x, y)
        ! in ssclib/wcs_aux

!print *, ''
!print *, 'theta, phi:', tanThetaArcsec(row), phi(row)
!print *, 'tanX, tanY:', tanX, tanY
!print *, '   x,    y:', x, y

        p_x(row) = nint(x)
        p_y(row) = nint(y)

        if ( x < 1 .or. x > imagesize*2&
        .or. y < 1 .or. y > imagesize*2) then
          numOutOfRange = numOutOfRange + 1
        endif

        attHi%ra   = max(attHi%ra,   scAttitude%ra)
        attLo%ra   = min(attLo%ra,   scAttitude%ra)
        attHi%dec  = max(attHi%dec,  scAttitude%dec)
        attLo%dec  = min(attLo%dec,  scAttitude%dec)
        attHi%apos = max(attHi%apos, scAttitude%apos)
        attLo%apos = min(attLo%apos, scAttitude%apos)
        Xmax = max(Xmax, x)
        Ymax = max(Ymax, y)
        Xmin = min(Xmin, x)
        Ymin = min(Ymin, y)
      end if
    end do
!---------------------------------------------------------
!
!*********************	END OF MAIN LOOP  ****************
!
!---------------------------------------------------------

    if (sourceOfAttInfo=='ODF' .and. writeBinnedAttPlease) then
      binnedAtt(numBins)%binEndTime = nearest(p_timeE(numEvents-1),1.0)
      binnedAttSetName = stringParameter('outbinnedattset')
      call writeBinnedAttitude(binnedAttSetName, binnedAtt(1:numBins)&
        , startTimeFirstBin, pntDeg, maxDelta, pntStyle, (/.false.,.false./))

      deallocate(binnedAtt)
    end if

    if (sourceOfAttInfo=='BINNED') deallocate(binnedAtt)

!---------------------------------------------------------
!	WRITE OUT A FEW USEFUL THINGS
!---------------------------------------------------------
    write(errstr,*) "No of input events:        ", numEvents
    call message(AppMsg,Verbose)

    if (sourceOfAttInfo /= "FIXED") then
      write(errstr,*) "No of event frames:        ", numTimeChanges
      call message(AppMsg,Verbose)
!      write(errstr,*) "No of bad attitude frames: ", numBadAtts
!      call message(AppMsg,Verbose)
      write(errstr,*) "No of bad attitude events: ", numBadEvents
      call message(AppMsg,Verbose)
      if (attHi%ra >= attLo%ra) then ! sensible ranges
        write(errstr,*) "Range in attitude RA:      ", attHi%ra/PiOn180&
          , attLo%ra/PiOn180
        call message(AppMsg,Verbose)
        write(errstr,*) "Range in attitude DEC:     ", attHi%dec/PiOn180&
          , attLo%dec/PiOn180
        call message(AppMsg,Verbose)
        write(errstr,*) "Range in attitude PA:      ", attHi%apos/PiOn180&
          , attLo%apos/PiOn180
        call message(AppMsg,Verbose)
      endif
    endif
    if (numEvents > 0) then 
      write(errstr,*)   "1st event X,Y:             ", p_x(0), p_y(0)
      call message(AppMsg,Noisy)
      if (numEvents > 3) then 
        write(errstr,*)   "2nd event X,Y:             ", p_x(1), p_y(1)
        call message(AppMsg,Noisy)
        write(errstr,*)   "3rd event X,Y:             ", p_x(2), p_y(2)
        call message(AppMsg,Noisy)
        write(errstr,*)   "4th event X,Y:             ", p_x(3), p_y(3)
        call message(AppMsg,Noisy)
      endif
    endif
    if (numOutOfRange > 0) then
      write(errstr,*) numOutOfRange," events ( of ",numEvents," ) out of &
        &range of image"
      call warning("outOfImageRange")
    endif
!--------------------------------------------
!  ADD WCS KEYWORDS
!--------------------------------------------
    call setAttribute(eventTab,"EQUINOX",2.000000E+03,""&
      ,"Equinox for sky coordinate x/y axes")
    call setAttribute(eventTab,"RADECSYS","FK5"&
      ,"World coord. system for this file")

    call setAttribute(xCol,"TLMIN",1,""&
      ,"minimum legal QPOE projected image X axis value")
    call setAttribute(xCol,"TLMAX",nint(2*imagesize),""&
      ,"maximum legal QPOE projected image X axis value")
    call setAttribute(xCol,"TDMIN",Xmin,""&
      ,"minimum projected image X axis data value")
    call setAttribute(xCol,"TDMAX",Xmax,""&
      ,"maximum projected image X axis data value")

    call setAttribute(yCol,"TLMIN",1,""&
      ,"minimum legal QPOE projected image Y axis value")
    call setAttribute(yCol,"TLMAX",nint(2*imagesize),""&
      ,"maximum legal QPOE projected image Y axis value")
    call setAttribute(yCol,"TDMIN",Ymin,""&
      ,"minimum projected image Y axis data value")
    call setAttribute(yCol,"TDMAX",Ymax,""&
      ,"maximum projected image Y axis data value")

    call writeWcs(xCol, yCol, xyColWcs) ! in ssclib/wcs_aux

!----------------------------------------------------------
!	ADD attcalc2 ATTRIBUTES TO EXTENSION
!----------------------------------------------------------
    call setAttribute(eventTab,"REFPOINT",tanRayStyle&
      ,"source of reference point information")
    call setAttribute(eventTab,"ATTSRC",sourceOfAttInfo&
      ,"source of attitude information")
    if (sourceOfAttInfo == "FIXED") then
      call setAttribute(eventTab,"ATTRA",scAttitude%ra/PiOn180,"deg"&
        ,"fixed attitude RA")
      call setAttribute(eventTab,"ATTDEC",scAttitude%dec/PiOn180,"deg"&
        ,"fixed attitude Dec")
      call setAttribute(eventTab,"ATTPA",scAttitude%apos/PiOn180,"deg"&
        ,"fixed attitude position angle")
    endif  

!--------------------------------------------------------------
!	ADD ATTRIBUTES TO PRIMARY EVENTS HEADER also WCS copies
!--------------------------------------------------------------
    call setAttribute(eventSet,"EQUINOX",2.000000E+03,""&
      ,"Equinox for sky coordinate x/y axes")
    call setAttribute(eventSet,"RADECSYS","FK5"&
      ,"World coord. system for this file")
    call setAttribute(eventSet,"REFXCTYP","RA---TAN"&
      ,"WCS Coord. type: RA tangent plane projection")
      value = nint(imagesize + 1)
    call setAttribute(eventSet,"REFXCRPX",value,""&
      ,"WCS axis reference pixel of projected image")
    call setAttribute(eventSet,"REFXCRVL",projRaDeg,"deg"&
      ,"WCS coord. at X axis ref. pixel")
!    call setAttribute(eventSet,"REFXCDLT",-pixsize_x,"deg/pix"&
    call setAttribute(eventSet,"REFXCDLT",-skyPixSizeDeg,"deg/pix"&
      ,"WCS X increment at ref. pixel")
    call setAttribute(eventSet,"REFXLMIN",1,""&
      ,"WCS minimum legal QPOE projected image X axis value")
      value = nint(2*imagesize)
    call setAttribute(eventSet,"REFXLMAX",value,""&
      ,"WCS maximum legal QPOE projected image X axis value")
    call setAttribute(eventSet,"REFXDMIN",Xmin,""&
      ,"WCS minimum projected image X axis data value")
    call setAttribute(eventSet,"REFXDMAX",Xmax,""&
      ,"WCS maximum projected image X axis data value")
    call setAttribute(eventSet,"REFXCUNI","deg"&
      ,"WCS Physical units of X axis")

    call setAttribute(eventSet,"REFYCTYP","DEC--TAN"&
      ,"WCS Coord. type: DEC tangent plane projection")
      value = nint(imagesize + 1)
    call setAttribute(eventSet,"REFYCRPX",value,""&
      ,"WCS axis reference pixel of projected image")
    call setAttribute(eventSet,"REFYCRVL",projDecDeg,"deg"&
      ,"WCS coord. at Y axis ref. pixel")
!    call setAttribute(eventSet,"REFYCDLT",pixsize_y,"deg/pix"&
    call setAttribute(eventSet,"REFYCDLT",skyPixSizeDeg,"deg/pix"&
      ,"WCS Y increment at ref. pixel")
    call setAttribute(eventSet,"REFYLMIN",1,""&
      ,"WCS minimum legal QPOE projected image Y axis value")
      value = nint(2*imagesize)
    call setAttribute(eventSet,"REFYLMAX",value,""&
      ,"WCS maximum legal QPOE projected image Y axis value")
    call setAttribute(eventSet,"REFYDMIN",Ymin,""&
      ,"WCS minimum projected image Y axis data value")
    call setAttribute(eventSet,"REFYDMAX",Ymax,""&
      ,"WCS maximum projected image Y axis data value")
    call setAttribute(eventSet,"REFYCUNI","deg"&
      ,"WCS Physical units of Y axis")

    call setAttribute(eventSet,"AVRG_PNT", trim(pntStyle)&
      ,"Meaning of _PNT values (mean, median or fixed)")

    call setAttribute(eventSet,"RA_PNT",pntDeg%ra,"deg"&
      ,"Average right ascension of the s/c pointing axis")
    call setAttribute(eventSet,"DEC_PNT",pntDeg%dec,"deg"&
      ,"Average declination of the s/c pointing axis")
    call setAttribute(eventSet,"PA_PNT",pntDeg%apos,"deg"&
      ,"Average position angle of the s/c pointing axis")

!-------------------------------------------------
!       ADD HISTORY (can put more information in)
!-------------------------------------------------
    history = "Source of tangent direction information: " // tanRayStyle
    call addHistory(eventSet,history)
    history = "Source of attitude information: " // sourceOfAttInfo
    call addHistory(eventSet,history)
    if (sourceOfAttInfo == "FIXED") then
      write(history,'(a,f13.7,a)') "Attitude RA = ",scAttitude%ra/PiOn180
      call addHistory(eventSet,history)
      write(history,'(a,f13.7,a)') "Attitude DEC = ",scAttitude%dec/PiOn180
      call addHistory(eventSet,history)
      write(history,'(a,f13.7,a)') "Attitude Pos.Angle = "&
        ,scAttitude%apos/PiOn180
      call addHistory(eventSet,history)
    endif  
!--------------------------------------------------
!	RELEASE EVENTS FILE
!--------------------------------------------------

!print *, 'WARNING! Test routine active.'
!p_x = p_x - xyColWcs%x%refPixels
!p_y = p_y - xyColWcs%y%refPixels
!print *, 'xref, yref:', xyColWcs%x%refPixels, xyColWcs%y%refPixels
    call release(eventSet)

    end subroutine evproject

  !---------------------------------------------------------------------
  subroutine nullcheck (thisNullable)
  !---------------------------------------------------------------------

    implicit none 

    type(NullableT),intent(in)  :: thisNullable

    !-------------------------------------------------------------------
    ! Start:
 
    if (.not.nullDefined(thisNullable)) then
      call warning("badXYNull","Null not defined correctly for X/Y columns")
    endif

  end subroutine nullcheck

  !---------------------------------------------------------------------
  subroutine template
  !---------------------------------------------------------------------

    implicit none 

    !-------------------------------------------------------------------
    ! Start:

  end subroutine template

end module evproject_mod

